#! /usr/local/bin/python3
# -*- coding: utf-8 -*-
##
## chat_sever.py: ein simpler Chat-Server
## Version 0.14, 2023-07-27
## W. Spiegel (walter.spiegel@web.de)
## Magic Key:
## Windows ==>  Strg + Untbr/Pause bzw. Ctrl + Break
## Linux   ==>  Strg + C   bzw. Ctrl + C
## To Do:
##      - Mitprotokollieren in Datei
##      - tkinter-Oberfläche
##      - Benutzer rauskicken
##      - Einstellungen
##
# stark verändert nach M. Lutz: Programming Python
# ch. 10, S. 547ff (Broadcast + locking hinzugefügt)
# with a little Help from J. Bishop: Java lernen, S. 524ff
# https://www.programcreek.com/python/example/83926/_thread.allocate_lock
# https://docs.python.org/3/library/_thread.html
# Vor dem Senden: Konvertierung nach bytes nur mit der Encoding-Angabe "utf-8"
# Nach dem Empfangen: Konvertierung zurück zu Strings nur mit der decode("utf-8") - Angabe
#########################################################
# Server side: open a socket on a port, listen for
# a message from a client, and send a broadcast; 
# echos lines until 'ENDE' when client closes socket;
# spawns a thread to handle each client connection;
# threads share global memory space with main thread;
#########################################################

import time, string, os
import _thread as thread
from socket import *        # get socket constructor and constants
myHost = ''                 # server machine, '' means local host
myPort = 8888               # listen on a non-reserved port number
nick_flag = 0               # 0: Adresse, 1: Nick-Name
nicki={}                    # enthält Adresse und Nick-Name
conn={}                     # enthält Adresse und connection-Parameter   
conn_lock = thread.allocate_lock()

sockobj = socket(AF_INET, SOCK_STREAM)  # make a TCP socket object
sockobj.bind((myHost, myPort))          # bind it to server port number
sockobj.listen(5)                       # allow upto 5 pending connects

if len(myHost) > 0:
    print( 'Server eingerichtet an Adresse',myHost,'auf Port',str(myPort) )
else:
    print( 'Server eingerichtet auf Port',str(myPort) )

if os.name == 'posix': 
    print( 'Magic Key: <Strg> + <C> [<Ctrl> + <C>]' )
else: 
    print( 'Magic Key: <Strg> + <Untbr> [<Ctrl> + <Break>]' )

#######################    
def now(): 
    return time.ctime(time.time())      # current time on the server
#######################    
def broadcast(msg, name):
    global conn
    conn_lock.acquire()
    chatter = conn.keys()
    for address in chatter:
        conn[address].send(bytes( str(name) +': '+ str(msg), "utf-8") )
    conn_lock.release()
#######################    
def handleClient(connection, address):  # in spawned thread: reply
    global nicki,conn
    
    connection.send( bytes( 'Verbunden! Beenden mit "ENDE", dein Nickname???', "utf-8") )
    data =  connection.recv(1024).decode("utf-8") 
    #print( type(connection.recv(1024)))
    if data:  ## doppelte Nicknamen verhindern!       
        while (data and data in nicki.values() or data.lower() in nicki.values() or data.upper() in nicki.values()):
            connection.send( bytes('Nickname "' + data + '" schon vergeben!!!! Denk dir einen neuen Namen aus!!', "utf-8") )
            data = connection.recv(1024).decode("utf-8")
            
    if data:        
        nicki[address]=data
        conn_lock.acquire()        
        conn[address]=connection
        conn_lock.release()        
        broadcast( str(data) + ' im Chatraum! ', '*** SERVER ***')    

    while 1:                        # read, write a client socket
        data =  connection.recv(1024).decode("utf-8") 
        #print( type(connection.recv(1024))) 
        if not data:
            break                   # keine Daten, das wars!
        if nick_flag:
            print( "Daten von '" + str(nicki.get(address) ) + "' erhalten: " + str(data) )
        else:
            print( "Daten von " + str(address) + " erhalten: " + str(data) )           

        if data != 'ENDE':
            broadcast (data,nicki.get(address) )
        else:  # data == 'ENDE', beenden!
            connection.send(b'OAIC') # magic string ciao rückwärts!
            conn_lock.acquire()
            del conn[address]
            conn_lock.release()            
            broadcast('Ciao ' + str(nicki.get(address)) ,'*** SERVER ***')
            del nicki[address]  
            break
        
    print( 'Beende Verbindung mit', address )
    connection.close()         

#######################    

def dispatcher():                                # listen until process killd
    while 1:                                     # wait for next connection,
        connection, address = sockobj.accept()   # pass to thread for service
        print( 'Server verbunden mit', address,',', now() ) 
        thread.start_new_thread(handleClient, (connection,address))

################# Hauptprogramm!
dispatcher()
